@@ -38,17 +38,17 @@ JackGraphManager* JackInternalClient::fGraphManager = NULL;
 JackEngineControl* JackInternalClient::fEngineControl = NULL;
 
 // Used for external C API (JackAPI.cpp)
-EXPORT JackGraphManager* GetGraphManager()
+SERVER_EXPORT JackGraphManager* GetGraphManager()
 {
     return JackServerGlobals::fInstance->GetGraphManager();
 }
 
-EXPORT JackEngineControl* GetEngineControl()
+SERVER_EXPORT JackEngineControl* GetEngineControl()
 {
     return JackServerGlobals::fInstance->GetEngineControl();
 }
 
-EXPORT JackSynchro* GetSynchroTable()
+SERVER_EXPORT JackSynchro* GetSynchroTable()
 {
     return JackServerGlobals::fInstance->GetSynchroTable();
 }
@@ -63,21 +63,31 @@ JackInternalClient::~JackInternalClient()
     delete fChannel;
 }
 
-int JackInternalClient::Open(const char* server_name, const char* name, int uuid, jack_options_t options, jack_status_t* status)
+int JackInternalClient::Open(const char* server_name, const char* name, jack_uuid_t uuid, jack_options_t options, jack_status_t* status)
 {
     int result;
-    char name_res[JACK_CLIENT_NAME_SIZE + 1];
     jack_log("JackInternalClient::Open name = %s", name);
+    
+    if (strlen(name) >= JACK_CLIENT_NAME_SIZE) {
+        jack_error("\"%s\" is too long to be used as a JACK client name.\n"
+                   "Please use %lu characters or less",
+                   name,
+                   JACK_CLIENT_NAME_SIZE - 1);
+        return -1; 
+    }
 
     strncpy(fServerName, server_name, sizeof(fServerName));
 
-    fChannel->ClientCheck(name, uuid, name_res, JACK_PROTOCOL_VERSION, (int)options, (int*)status, &result);
+    // Open server/client direct channel
+    char name_res[JACK_CLIENT_NAME_SIZE + 1];
+    fChannel->ClientCheck(name, uuid, name_res, JACK_PROTOCOL_VERSION, (int)options, (int*)status, &result, false);
     if (result < 0) {
         int status1 = *status;
-        if (status1 & JackVersionError)
+        if (status1 & JackVersionError) {
             jack_error("JACK protocol mismatch %d", JACK_PROTOCOL_VERSION);
-        else
+        } else {
             jack_error("Client name = %s conflits with another running client", name);
+        }
         goto error;
     }
 
@@ -97,11 +107,16 @@ int JackInternalClient::Open(const char* server_name, const char* name, int uuid
     return 0;
 
 error:
-    fChannel->Stop();
     fChannel->Close();
     return -1;
 }
 
+void JackInternalClient::ShutDown(jack_status_t code, const char* message)
+{
+    jack_log("JackInternalClient::ShutDown");
+    JackClient::ShutDown(code, message);
+}
+
 JackGraphManager* JackInternalClient::GetGraphManager() const
 {
     assert(fGraphManager);
@@ -181,7 +196,10 @@ int JackLoadableInternalClient2::Init(const char* so_name)
 JackLoadableInternalClient1::JackLoadableInternalClient1(JackServer* server, JackSynchro* table, const char* object_data)
         : JackLoadableInternalClient(server, table)
 {
-    strncpy(fObjectData, object_data, JACK_LOAD_INIT_LIMIT);
+    if (object_data != NULL)
+        strncpy(fObjectData, object_data, JACK_LOAD_INIT_LIMIT);
+    else
+        memset(fObjectData, 0, sizeof(fObjectData));
 }
 
 JackLoadableInternalClient2::JackLoadableInternalClient2(JackServer* server, JackSynchro* table, const JSList*  parameters)
@@ -192,13 +210,15 @@ JackLoadableInternalClient2::JackLoadableInternalClient2(JackServer* server, Jac
 
 JackLoadableInternalClient::~JackLoadableInternalClient()
 {
-    if (fFinish != NULL)
+    if (fFinish != NULL) {
         fFinish(fProcessArg);
-    if (fHandle != NULL)
+    }
+    if (fHandle != NULL) {
         UnloadJackModule(fHandle);
+    }
 }
 
-int JackLoadableInternalClient1::Open(const char* server_name, const char* name, int uuid, jack_options_t options, jack_status_t* status)
+int JackLoadableInternalClient1::Open(const char* server_name, const char* name, jack_uuid_t uuid, jack_options_t options, jack_status_t* status)
 {
     int res = -1;
 
@@ -214,7 +234,7 @@ int JackLoadableInternalClient1::Open(const char* server_name, const char* name,
     return res;
 }
 
-int JackLoadableInternalClient2::Open(const char* server_name, const char* name, int uuid, jack_options_t options, jack_status_t* status)
+int JackLoadableInternalClient2::Open(const char* server_name, const char* name, jack_uuid_t uuid, jack_options_t options, jack_status_t* status)
 {
     int res = -1;
 
