@@ -5,6 +5,7 @@
  *	- common (interface-independent) code
  *	- POSIX implementation
  *	- System V implementation
+ *  - Windows implementation
  *
  * The implementation used is determined by whether USE_POSIX_SHM was
  * set in the ./configure step.
@@ -12,21 +13,22 @@
 
 /*
  Copyright (C) 2001-2003 Paul Davis
+ Copyright (C) 2005-2012 Grame
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU Lesser General Public License as published by
  the Free Software Foundation; either version 2.1 of the License, or
  (at your option) any later version.
- 
+
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU Lesser General Public License for more details.
- 
+
  You should have received a copy of the GNU Lesser General Public License
- along with this program; if not, write to the Free Software 
+ along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
- 
+
  */
 
 #include "JackConstants.h"
@@ -50,6 +52,7 @@
 #include <sys/shm.h>
 #include <sys/sem.h>
 #include <stdlib.h>
+#include "promiscuous.h"
 
 #endif
 
@@ -62,7 +65,7 @@ static int GetUID()
     return  _getpid();
     //#error "No getuid function available"
 #else
-    return getuid();
+    return geteuid();
 #endif
 }
 
@@ -117,7 +120,7 @@ static jack_shm_info_t registry_info = { /* SHM info for the registry */
 /* pointers to registry header and array */
 static jack_shm_header_t   *jack_shm_header = NULL;
 static jack_shm_registry_t *jack_shm_registry = NULL;
-static char jack_shm_server_prefix[JACK_SERVER_NAME_SIZE] = "";
+static char jack_shm_server_prefix[JACK_SERVER_NAME_SIZE+1] = "";
 
 /* jack_shm_lock_registry() serializes updates to the shared memory
  * segment JACK uses to keep track of the SHM segments allocated to
@@ -143,6 +146,31 @@ static int semid = -1;
 
 #ifdef WIN32
 
+#include <psapi.h>
+#include <lmcons.h>
+
+static BOOL check_process_running(DWORD process_id)
+{
+    DWORD aProcesses[2048], cbNeeded, cProcesses;
+    unsigned int i;
+
+    // Enumerate all processes
+    if (!EnumProcesses(aProcesses, sizeof(aProcesses), &cbNeeded)) {
+        return FALSE;
+    }
+
+    // Calculate how many process identifiers were returned.
+    cProcesses = cbNeeded / sizeof(DWORD);
+
+    for (i = 0; i < cProcesses; i++) {
+         if (aProcesses[i] == process_id) {
+            // Process process_id is running...
+            return TRUE;
+        }
+    }
+    return FALSE;
+}
+
 static int
 semaphore_init () {return 0;}
 
@@ -289,8 +317,16 @@ jack_shm_validate_registry ()
 static void
 jack_set_server_prefix (const char *server_name)
 {
-	snprintf (jack_shm_server_prefix, sizeof (jack_shm_server_prefix),
+#ifdef WIN32
+    char buffer[UNLEN+1]={0};
+    DWORD len = UNLEN+1;
+    GetUserName(buffer, &len);
+    snprintf (jack_shm_server_prefix, sizeof (jack_shm_server_prefix),
+		  "jack-%s:%s:", buffer, server_name);
+#else
+    snprintf (jack_shm_server_prefix, sizeof (jack_shm_server_prefix),
 		  "jack-%d:%s:", GetUID(), server_name);
+#endif
 }
 
 /* gain server addressability to shared memory registration segment
@@ -481,7 +517,12 @@ jack_register_server (const char *server_name, int new_registry)
 		}
 
 		/* see if server still exists */
-	#ifndef WIN32 // steph TO CHECK
+    #ifdef WIN32
+        if (check_process_running(jack_shm_header->server[i].pid)) {
+            res = EEXIST;	/* other server running */
+			goto unlock;
+        }
+	#else
 		if (kill (jack_shm_header->server[i].pid, 0) == 0)  {
 			res = EEXIST;	/* other server running */
 			goto unlock;
@@ -571,8 +612,8 @@ jack_cleanup_shm ()
 		} else {
 
 			/* see if allocator still exists */
-		#ifdef WIN32 // steph
-			jack_info("TODO: kill API not available !!");
+		#ifdef WIN32 
+			//jack_info("TODO: kill API not available !!");
 		#else
 			if (kill (r->allocator, 0)) {
 				if (errno == ESRCH) {
@@ -614,7 +655,7 @@ jack_cleanup_shm ()
  * from here.
  *
  * This is not done under a single lock.  I don't even want to think
- * about all the things that could possibly go wrong if multple
+ * about all the things that could possibly go wrong if multiple
  * processes tried to resize the same segment concurrently.  That
  * probably doesn't happen.
  */
@@ -636,6 +677,24 @@ jack_resize_shm (jack_shm_info_t* si, jack_shmsize_t size)
 	return jack_attach_shm (si);
 }
 
+int
+jack_attach_lib_shm (jack_shm_info_t* si)
+{
+    int res = jack_attach_shm(si);
+    if (res == 0)
+        si->size = jack_shm_registry[si->index].size; // Keep size in si struct
+    return res;
+}
+
+int
+jack_attach_lib_shm_read (jack_shm_info_t* si)
+{
+    int res = jack_attach_shm_read(si);
+    if (res == 0)
+        si->size = jack_shm_registry[si->index].size; // Keep size in si struct
+    return res;
+}
+
 #ifdef USE_POSIX_SHM
 
 /* * * * * * * * * * * * * * * * * * * * * * * * * * * * * * *
@@ -683,7 +742,7 @@ jack_access_registry (jack_shm_info_t *ri)
 	jack_shm_header = ri->ptr.attached_at;
 	jack_shm_registry = (jack_shm_registry_t *) (jack_shm_header + 1);
 
-	close (shm_fd); // steph
+	close (shm_fd); 
 	return 0;
 }
 
@@ -709,7 +768,7 @@ jack_create_registry (jack_shm_info_t *ri)
 		return rc;
 	}
 
-    /* Previous shm_open result depends of the actual value of umask, force correct file permisssion here */
+    /* Previous shm_open result depends of the actual value of umask, force correct file permission here */
     if (fchmod(shm_fd, S_IRUSR | S_IWUSR | S_IRGRP | S_IWGRP | S_IROTH | S_IWOTH) < 0) {
 	    jack_log("Cannot chmod jack-shm-registry (%s) %d %d", strerror (errno));
     }
@@ -742,7 +801,7 @@ jack_create_registry (jack_shm_info_t *ri)
 
 	/* initialize registry contents */
 	jack_shm_init_registry ();
-	close (shm_fd); // steph
+	close (shm_fd); 
 	return 0;
 }
 
@@ -759,6 +818,15 @@ jack_release_shm (jack_shm_info_t* si)
 	/* registry may or may not be locked */
 	if (si->ptr.attached_at != MAP_FAILED) {
 		munmap (si->ptr.attached_at, jack_shm_registry[si->index].size);
+  	}
+}
+
+void
+jack_release_lib_shm (jack_shm_info_t* si)
+{
+	/* registry may or may not be locked */
+	if (si->ptr.attached_at != MAP_FAILED) {
+       munmap (si->ptr.attached_at, si->size);
 	}
 }
 
@@ -770,6 +838,7 @@ jack_shmalloc (const char *shm_name, jack_shmsize_t size, jack_shm_info_t* si)
 	int shm_fd;
 	int rc = -1;
 	char name[SHM_NAME_MAX+1];
+	const char* promiscuous;
 
 	if (jack_shm_lock_registry () < 0) {
         jack_error ("jack_shm_lock_registry fails...");
@@ -810,6 +879,10 @@ jack_shmalloc (const char *shm_name, jack_shmsize_t size, jack_shm_info_t* si)
 		goto unlock;
 	}
 
+	promiscuous = getenv("JACK_PROMISCUOUS_SERVER");
+	if ((promiscuous != NULL) && (jack_promiscuous_perms(shm_fd, name, jack_group2gid(promiscuous)) < 0))
+		goto unlock;
+
 	close (shm_fd);
 	registry->size = size;
 	strncpy (registry->id, name, sizeof (registry->id));
@@ -961,6 +1034,12 @@ jack_release_shm (jack_shm_info_t* si)
 	}
 }
 
+void
+jack_release_lib_shm (jack_shm_info_t* si)
+{
+	jack_release_shm(si);
+}
+
 int
 jack_shmalloc (const char *shm_name, jack_shmsize_t size, jack_shm_info_t* si)
 {
@@ -1096,7 +1175,7 @@ jack_access_registry (jack_shm_info_t *ri)
 		}
 	}
 
-	if ((ri->attached_at = shmat (registry_id, 0, 0)) < 0) {
+	if ((ri->ptr.attached_at = shmat (registry_id, 0, 0)) < 0) {
 		jack_error ("Cannot attach shm registry segment (%s)",
 			    strerror (errno));
 		return EINVAL;
@@ -1104,7 +1183,7 @@ jack_access_registry (jack_shm_info_t *ri)
 
 	/* set up global pointers */
 	ri->index = JACK_SHM_REGISTRY_INDEX;
-	jack_shm_header = ri->attached_at;
+	jack_shm_header = ri->ptr.attached_at;
 	jack_shm_registry = (jack_shm_registry_t *) (jack_shm_header + 1);
 	return 0;
 }
@@ -1128,7 +1207,7 @@ jack_create_registry (jack_shm_info_t *ri)
 		return errno;
 	}
 
-	if ((ri->attached_at = shmat (registry_id, 0, 0)) < 0) {
+	if ((ri->ptr.attached_at = shmat (registry_id, 0, 0)) < 0) {
 		jack_error ("Cannot attach shm registry segment (%s)",
 			    strerror (errno));
 		return EINVAL;
@@ -1136,7 +1215,7 @@ jack_create_registry (jack_shm_info_t *ri)
 
 	/* set up global pointers */
 	ri->index = JACK_SHM_REGISTRY_INDEX;
-	jack_shm_header = ri->attached_at;
+	jack_shm_header = ri->ptr.attached_at;
 	jack_shm_registry = (jack_shm_registry_t *) (jack_shm_header + 1);
 
 	/* initialize registry contents */
@@ -1155,11 +1234,17 @@ void
 jack_release_shm (jack_shm_info_t* si)
 {
 	/* registry may or may not be locked */
-	if (si->attached_at != MAP_FAILED) {
-		shmdt (si->attached_at);
+	if (si->ptr.attached_at != MAP_FAILED) {
+		shmdt (si->ptr.attached_at);
 	}
 }
 
+void
+jack_release_lib_shm (jack_shm_info_t* si)
+{
+	jack_release_shm(si);
+}
+
 int
 jack_shmalloc (const char* name_not_used, jack_shmsize_t size,
 	       jack_shm_info_t* si)
@@ -1184,7 +1269,7 @@ jack_shmalloc (const char* name_not_used, jack_shmsize_t size,
 			registry->id = shmid;
 			registry->allocator = getpid();
 			si->index = registry->index;
-			si->attached_at = MAP_FAILED; /* not attached */
+			si->ptr.attached_at = MAP_FAILED; /* not attached */
 			rc = 0;
 
 		} else {
@@ -1200,7 +1285,19 @@ jack_shmalloc (const char* name_not_used, jack_shmsize_t size,
 int
 jack_attach_shm (jack_shm_info_t* si)
 {
-	if ((si->attached_at = shmat (jack_shm_registry[si->index].id, 0, 0)) < 0) {
+	if ((si->ptr.attached_at = shmat (jack_shm_registry[si->index].id, 0, 0)) < 0) {
+		jack_error ("Cannot attach shm segment (%s)",
+			    strerror (errno));
+		jack_release_shm_info (si->index);
+		return -1;
+	}
+	return 0;
+}
+
+int
+jack_attach_shm_read (jack_shm_info_t* si)
+{
+	if ((si->ptr.attached_at = shmat (jack_shm_registry[si->index].id, 0, SHM_RDONLY)) < 0) {
 		jack_error ("Cannot attach shm segment (%s)",
 			    strerror (errno));
 		jack_release_shm_info (si->index);
@@ -1210,3 +1307,4 @@ jack_attach_shm (jack_shm_info_t* si)
 }
 
 #endif /* !USE_POSIX_SHM */
+
