@@ -20,6 +20,7 @@
 
 
 #define __STDC_FORMAT_MACROS   // For inttypes.h to work in C++
+#define _GNU_SOURCE            /* for strcasestr() from string.h */
 
 #include <math.h>
 #include <stdio.h>
@@ -31,7 +32,6 @@
 #include <signal.h>
 #include <sys/types.h>
 #include <sys/time.h>
-#include <regex.h>
 #include <string.h>
 
 #include "alsa_driver.h"
@@ -52,8 +52,12 @@ extern void show_work_times ();
 
 #undef DEBUG_WAKEUP
 
+char* strcasestr(const char* haystack, const char* needle);
+
 /* Delay (in process calls) before jackd will report an xrun */
 #define XRUN_REPORT_DELAY 0
+/* Max re-try count for Alsa poll timeout handling */
+#define MAX_RETRY_COUNT 5
 
 void
 jack_driver_init (jack_driver_t *driver)
@@ -134,30 +138,18 @@ alsa_driver_check_capabilities (alsa_driver_t *driver)
 	return 0;
 }
 
+char* get_control_device_name(const char * device_name);
+
 static int
 alsa_driver_check_card_type (alsa_driver_t *driver)
 {
 	int err;
 	snd_ctl_card_info_t *card_info;
 	char * ctl_name;
-	regex_t expression;
 
 	snd_ctl_card_info_alloca (&card_info);
 
-	regcomp(&expression,"(plug)?hw:[0-9](,[0-9])?",REG_ICASE|REG_EXTENDED);
-
-       	if (!regexec(&expression,driver->alsa_name_playback,0,NULL,0)) {
-		/* the user wants a hw or plughw device, the ctl name
-		 * should be hw:x where x is the card number */
-
-		char tmp[5];
-		strncpy(tmp,strstr(driver->alsa_name_playback,"hw"),4);
-		tmp[4]='\0';
-		jack_info("control device %s",tmp);
-		ctl_name = strdup(tmp);
-	} else {
-		ctl_name = strdup(driver->alsa_name_playback);
-	}
+	ctl_name = get_control_device_name(driver->alsa_name_playback);
 
 	// XXX: I don't know the "right" way to do this. Which to use
 	// driver->alsa_name_playback or driver->alsa_name_capture.
@@ -172,7 +164,6 @@ alsa_driver_check_card_type (alsa_driver_t *driver)
 
 	driver->alsa_driver = strdup(snd_ctl_card_info_get_driver (card_info));
 
-	regfree(&expression);
 	free(ctl_name);
 
 	return alsa_driver_check_capabilities (driver);
@@ -204,8 +195,8 @@ alsa_driver_ice1712_hardware (alsa_driver_t *driver)
 static int
 alsa_driver_usx2y_hardware (alsa_driver_t *driver)
 {
-        driver->hw = jack_alsa_usx2y_hw_new (driver);
-	return 0;
+    driver->hw = jack_alsa_usx2y_hw_new (driver);
+    return 0;
 }
 */
 
@@ -375,15 +366,17 @@ alsa_driver_configure_stream (alsa_driver_t *driver, char *device_name,
 	unsigned int frame_rate;
 	snd_pcm_uframes_t stop_th;
 	static struct {
-		char Name[32];
+		char Name[40];
 		snd_pcm_format_t format;
 		int swapped;
 	} formats[] = {
- 	    {"32bit float little-endian", SND_PCM_FORMAT_FLOAT_LE},
+		{"32bit float little-endian", SND_PCM_FORMAT_FLOAT_LE, IS_LE},
 		{"32bit integer little-endian", SND_PCM_FORMAT_S32_LE, IS_LE},
 		{"32bit integer big-endian", SND_PCM_FORMAT_S32_BE, IS_BE},
-		{"24bit little-endian", SND_PCM_FORMAT_S24_3LE, IS_LE},
-		{"24bit big-endian", SND_PCM_FORMAT_S24_3BE, IS_BE},
+		{"24bit little-endian in 3bytes format", SND_PCM_FORMAT_S24_3LE, IS_LE},
+		{"24bit big-endian in 3bytes format", SND_PCM_FORMAT_S24_3BE, IS_BE},
+		{"24bit little-endian", SND_PCM_FORMAT_S24_LE, IS_LE},
+		{"24bit big-endian", SND_PCM_FORMAT_S24_BE, IS_BE},
 		{"16bit little-endian", SND_PCM_FORMAT_S16_LE, IS_LE},
 		{"16bit big-endian", SND_PCM_FORMAT_S16_BE, IS_BE},
 	};
@@ -598,6 +591,20 @@ alsa_driver_configure_stream (alsa_driver_t *driver, char *device_name,
 		return -1;
 	}
 
+	err = snd_pcm_sw_params_set_tstamp_mode(handle, sw_params, SND_PCM_TSTAMP_ENABLE);
+	if (err < 0) {
+		jack_info("Could not enable ALSA time stamp mode for %s (err %d)",
+			  stream_name, err);
+	}
+
+#if SND_LIB_MAJOR >= 1 && SND_LIB_MINOR >= 1
+	err = snd_pcm_sw_params_set_tstamp_type(handle, sw_params, SND_PCM_TSTAMP_TYPE_MONOTONIC);
+	if (err < 0) {
+		jack_info("Could not use monotonic ALSA time stamps for %s (err %d)",
+			  stream_name, err);
+	}
+#endif
+
 	if ((err = snd_pcm_sw_params (handle, sw_params)) < 0) {
 		jack_error ("ALSA: cannot set software parameters for %s\n",
 			    stream_name);
@@ -661,7 +668,7 @@ alsa_driver_set_parameters (alsa_driver_t *driver,
 		}
 	}
 
-	/* check the rate, since thats rather important */
+	/* check the rate, since that's rather important */
 
 	if (driver->playback_handle) {
 		snd_pcm_hw_params_get_rate (driver->playback_hw_params,
@@ -706,7 +713,7 @@ alsa_driver_set_parameters (alsa_driver_t *driver,
 	}
 
 
-	/* check the fragment size, since thats non-negotiable */
+	/* check the fragment size, since that's non-negotiable */
 
 	if (driver->playback_handle) {
  		snd_pcm_access_t access;
@@ -767,6 +774,8 @@ alsa_driver_set_parameters (alsa_driver_t *driver,
 		case SND_PCM_FORMAT_S32_LE:
 		case SND_PCM_FORMAT_S24_3LE:
 		case SND_PCM_FORMAT_S24_3BE:
+		case SND_PCM_FORMAT_S24_LE:
+		case SND_PCM_FORMAT_S24_BE:
 		case SND_PCM_FORMAT_S16_LE:
 		case SND_PCM_FORMAT_S32_BE:
 		case SND_PCM_FORMAT_S16_BE:
@@ -785,6 +794,8 @@ alsa_driver_set_parameters (alsa_driver_t *driver,
 		case SND_PCM_FORMAT_S32_LE:
 		case SND_PCM_FORMAT_S24_3LE:
 		case SND_PCM_FORMAT_S24_3BE:
+		case SND_PCM_FORMAT_S24_LE:
+		case SND_PCM_FORMAT_S24_BE:
 		case SND_PCM_FORMAT_S16_LE:
 		case SND_PCM_FORMAT_S32_BE:
 		case SND_PCM_FORMAT_S16_BE:
@@ -903,6 +914,9 @@ alsa_driver_set_parameters (alsa_driver_t *driver,
 */
 
 	return 0;
+
+	// may be unused
+	(void)err;
 }
 
 int
@@ -925,7 +939,7 @@ alsa_driver_get_channel_addresses (alsa_driver_t *driver,
 				   snd_pcm_uframes_t *capture_offset,
 				   snd_pcm_uframes_t *playback_offset)
 {
-	unsigned long err;
+	int err;
 	channel_t chn;
 
 	if (capture_avail) {
@@ -1201,7 +1215,8 @@ alsa_driver_xrun_recovery (alsa_driver_t *driver, float *delayed_usecs)
 			    < 0) {
 				jack_error("error preparing after suspend: %s", snd_strerror(res));
 			}
-		} else {
+		} 
+		if (driver->playback_handle) {
 			if ((res = snd_pcm_prepare(driver->playback_handle))
 			    < 0) {
 				jack_error("error preparing after suspend: %s", snd_strerror(res));
@@ -1218,6 +1233,18 @@ alsa_driver_xrun_recovery (alsa_driver_t *driver, float *delayed_usecs)
 		timersub(&now, &tstamp, &diff);
 		*delayed_usecs = diff.tv_sec * 1000000.0 + diff.tv_usec;
 		jack_log("**** alsa_pcm: xrun of at least %.3f msecs",*delayed_usecs / 1000.0);
+		if (driver->capture_handle) {
+			jack_log("Repreparing capture");
+			if ((res = snd_pcm_prepare(driver->capture_handle)) < 0) {
+				jack_error("error preparing after xrun: %s", snd_strerror(res));
+			}
+		}
+		if (driver->playback_handle) {
+			jack_log("Repreparing playback");
+			if ((res = snd_pcm_prepare(driver->playback_handle)) < 0) {
+				jack_error("error preparing after xrun: %s", snd_strerror(res));
+			}
+		}
 	}
 
 	if (alsa_driver_restart (driver)) {
@@ -1265,6 +1292,7 @@ alsa_driver_wait (alsa_driver_t *driver, int extra_fd, int *status, float
 	int xrun_detected = FALSE;
 	int need_capture;
 	int need_playback;
+	int retry_cnt = 0;
 	unsigned int i;
 	jack_time_t poll_enter;
 	jack_time_t poll_ret = 0;
@@ -1282,7 +1310,7 @@ alsa_driver_wait (alsa_driver_t *driver, int extra_fd, int *status, float
 
   again:
 
-	while (need_playback || need_capture) {
+	while ((need_playback || need_capture) && !xrun_detected) {
 
 		int poll_result;
 		unsigned int ci = 0;
@@ -1331,7 +1359,11 @@ alsa_driver_wait (alsa_driver_t *driver, int extra_fd, int *status, float
 			driver->poll_late++;
 		}
 
+#ifdef __ANDROID__
+		poll_result = poll (driver->pfd, nfds, -1);  //fix for sleep issue
+#else
 		poll_result = poll (driver->pfd, nfds, driver->poll_timeout);
+#endif
 		if (poll_result < 0) {
 
 			if (errno == EINTR) {
@@ -1354,6 +1386,25 @@ alsa_driver_wait (alsa_driver_t *driver, int extra_fd, int *status, float
 
 		poll_ret = jack_get_microseconds ();
 
+		if (poll_result == 0) {
+			retry_cnt++;
+			if(retry_cnt > MAX_RETRY_COUNT) {
+				jack_error ("ALSA: poll time out, polled for %" PRIu64
+					    " usecs, Reached max retry cnt = %d, Exiting",
+					    poll_ret - poll_enter, MAX_RETRY_COUNT);
+				*status = -5;
+				return 0;
+			}
+			jack_log ("ALSA: poll time out, polled for %" PRIu64
+				  " usecs, Retrying with a recovery, retry cnt = %d",
+				  poll_ret - poll_enter, retry_cnt);
+			*status = alsa_driver_xrun_recovery (driver, delayed_usecs);
+			if(*status != 0) {
+				jack_error ("ALSA: poll time out, recovery failed with status = %d", *status);
+				return 0;
+			}
+		}
+
         // JACK2
         SetTime(poll_ret);
 
@@ -1402,6 +1453,12 @@ alsa_driver_wait (alsa_driver_t *driver, int extra_fd, int *status, float
 				return 0;
 			}
 
+			if (revents & POLLNVAL) {
+				jack_error ("ALSA: playback device disconnected");
+				*status = -7;
+				return 0;
+			}
+
 			if (revents & POLLERR) {
 				xrun_detected = TRUE;
 			}
@@ -1425,6 +1482,12 @@ alsa_driver_wait (alsa_driver_t *driver, int extra_fd, int *status, float
 				return 0;
 			}
 
+			if (revents & POLLNVAL) {
+				jack_error ("ALSA: capture device disconnected");
+				*status = -7;
+				return 0;
+			}
+
 			if (revents & POLLERR) {
 				xrun_detected = TRUE;
 			}
@@ -1438,15 +1501,6 @@ alsa_driver_wait (alsa_driver_t *driver, int extra_fd, int *status, float
 #endif
 			}
 		}
-
-		if (poll_result == 0) {
-			jack_error ("ALSA: poll time out, polled for %" PRIu64
-				    " usecs",
-				    poll_ret - poll_enter);
-			*status = -5;
-			return 0;
-		}
-
 	}
 
 	if (driver->capture_handle) {
@@ -1823,6 +1877,107 @@ alsa_driver_delete (alsa_driver_t *driver)
 	free (driver);
 }
 
+static char*
+discover_alsa_using_apps ()
+{
+        char found[2048];
+        char command[5192];
+        char* path = getenv ("PATH");
+        char* dir;
+        size_t flen = 0;
+        int card;
+        int device;
+        size_t cmdlen = 0;
+
+        if (!path) {
+                return NULL;
+        }
+
+        /* look for lsof and give up if its not in PATH */
+
+        path = strdup (path);
+        dir = strtok (path, ":");
+        while (dir) {
+                char maybe[PATH_MAX+1];
+                snprintf (maybe, sizeof(maybe), "%s/lsof", dir);
+                if (access (maybe, X_OK) == 0) {
+                        break;
+                }
+                dir = strtok (NULL, ":");
+        }
+        free (path);
+
+        if (!dir) {
+                return NULL;
+        }
+
+        snprintf (command, sizeof (command), "lsof -Fc0 ");
+        cmdlen = strlen (command);
+
+        for (card = 0; card < 8; ++card) {
+                for (device = 0; device < 8; ++device)  {
+                        char buf[32];
+
+                        snprintf (buf, sizeof (buf), "/dev/snd/pcmC%dD%dp", card, device);
+                        if (access (buf, F_OK) == 0) {
+                                snprintf (command+cmdlen, sizeof(command)-cmdlen, "%s ", buf);
+                        }
+                        cmdlen = strlen (command);
+
+                        snprintf (buf, sizeof (buf), "/dev/snd/pcmC%dD%dc", card, device);
+                        if (access (buf, F_OK) == 0) {
+                                snprintf (command+cmdlen, sizeof(command)-cmdlen, "%s ", buf);
+                        }
+                        cmdlen = strlen (command);
+                }
+        }
+
+        FILE* f = popen (command, "r");
+
+        if (!f) {
+                return NULL;
+        }
+
+        while (!feof (f)) {
+                char buf[1024]; /* lsof doesn't output much */
+
+                if (!fgets (buf, sizeof (buf), f)) {
+                        break;
+                }
+
+                if (*buf != 'p') {
+                        return NULL;
+                }
+
+                /* buf contains NULL as a separator between the process field and the command field */
+                char *pid = buf;
+                ++pid; /* skip leading 'p' */
+                char *cmd = pid;
+
+                /* skip to NULL */
+                while (*cmd) {
+                        ++cmd;
+                }
+                ++cmd; /* skip to 'c' */
+                ++cmd; /* skip to first character of command */
+
+                snprintf (found+flen, sizeof (found)-flen, "%s (process ID %s)\n", cmd, pid);
+                flen = strlen (found);
+
+                if (flen >= sizeof (found)) {
+                        break;
+                }
+        }
+
+        pclose (f);
+
+        if (flen) {
+                return strdup (found);
+        } else {
+                return NULL;
+        }
+}
+
 jack_driver_t *
 alsa_driver_new (char *name, char *playback_alsa_device,
 		 char *capture_alsa_device,
@@ -1846,7 +2001,7 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 		 )
 {
 	int err;
-
+    char* current_apps;
 	alsa_driver_t *driver;
 
 	jack_info ("creating alsa driver ... %s|%s|%" PRIu32 "|%" PRIu32
@@ -1944,14 +2099,33 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 				  SND_PCM_NONBLOCK) < 0) {
 			switch (errno) {
 			case EBUSY:
-				jack_error ("the playback device \"%s\" is "
-					    "already in use. Please stop the"
-					    " application using it and "
-					    "run JACK again",
-					    playback_alsa_device);
-				alsa_driver_delete (driver);
+#ifdef __ANDROID__
+                jack_error ("\n\nATTENTION: The playback device \"%s\" is "
+                            "already in use. Please stop the"
+                            " application using it and "
+                            "run JACK again",
+                            playback_alsa_device);
+#else
+                current_apps = discover_alsa_using_apps ();
+                if (current_apps) {
+                        jack_error ("\n\nATTENTION: The playback device \"%s\" is "
+                                    "already in use. The following applications "
+                                    " are using your soundcard(s) so you should "
+                                    " check them and stop them as necessary before "
+                                    " trying to start JACK again:\n\n%s",
+                                    playback_alsa_device,
+                                    current_apps);
+                        free (current_apps);
+                } else {
+                        jack_error ("\n\nATTENTION: The playback device \"%s\" is "
+                                    "already in use. Please stop the"
+                                    " application using it and "
+                                    "run JACK again",
+                                    playback_alsa_device);
+                }
+#endif
+                alsa_driver_delete (driver);
 				return NULL;
-				break;
 
 			case EPERM:
 				jack_error ("you do not have permission to open "
@@ -1977,13 +2151,31 @@ alsa_driver_new (char *name, char *playback_alsa_device,
 				  SND_PCM_NONBLOCK) < 0) {
 			switch (errno) {
 			case EBUSY:
-				jack_error ("the capture device \"%s\" is "
-					    "already in use. Please stop the"
-					    " application using it and "
-					    "run JACK again",
-					    capture_alsa_device);
+#ifdef __ANDROID__
+                jack_error ("\n\nATTENTION: The capture (recording) device \"%s\" is "
+                            "already in use",
+                            capture_alsa_device);
+#else
+                current_apps = discover_alsa_using_apps ();
+                if (current_apps) {
+                        jack_error ("\n\nATTENTION: The capture device \"%s\" is "
+                                    "already in use. The following applications "
+                                    " are using your soundcard(s) so you should "
+                                    " check them and stop them as necessary before "
+                                    " trying to start JACK again:\n\n%s",
+                                    capture_alsa_device,
+                                    current_apps);
+                        free (current_apps);
+                } else {
+                        jack_error ("\n\nATTENTION: The capture (recording) device \"%s\" is "
+                                    "already in use. Please stop the"
+                                    " application using it and "
+                                    "run JACK again",
+                                    capture_alsa_device);
+                }
 				alsa_driver_delete (driver);
 				return NULL;
+#endif
 				break;
 
 			case EPERM:
